'use strict';

var streamingIterables = require('streaming-iterables');
var core = require('@xobj/core');
var queueable = require('queueable');
var pino = require('pino');

function _interopDefault (e) { return e && e.__esModule ? e : { default: e }; }

var pino__default = /*#__PURE__*/_interopDefault(pino);

// src/client.ts
var defaultCodec = {
  decoder: (array) => core.decode(
    array.buffer.slice(array.byteOffset, array.byteLength + array.byteOffset)
  ),
  encoder: (data) => new Uint8Array(core.encode(data))
};

// src/common.ts
var defaultInitOptions = {
  codec: defaultCodec,
  context: {}
};
var control_name = "libp2p-transport/control";
var logger = pino__default.default();

// src/utils.ts
var newChannel = (c, i, ctrl, context) => {
  const id = { connection: i.connection.id, stream: i.stream.id };
  let open = true;
  return {
    send: async (value) => {
      if (!open) {
        throw "channel has already closed";
      }
      await c.push(value);
      logger.trace(`Send ${JSON.stringify(value)} to ${JSON.stringify(id)}`);
    },
    done: async (err) => {
      await c.return();
      if (ctrl) {
        if (err) {
          await ctrl.push({
            type: "error",
            id,
            name: err.name,
            message: err.message,
            stack: void 0
          });
          logger.debug(`${JSON.stringify(id)} chan done with error`);
        } else {
          await ctrl.push({
            type: "success",
            id
          });
          logger.trace(`${JSON.stringify(id)} chan done with success`);
        }
      }
      open = false;
    },
    id,
    protocol: i.stream.protocol,
    context,
    stat: {
      remoteAddr: i.connection.remoteAddr,
      remotePeer: i.connection.remotePeer,
      encryption: i.connection.encryption,
      multiplexer: i.connection.multiplexer,
      direction: i.stream.direction,
      open: () => open
    }
  };
};
var makeNext = (defaultParams) => async (params) => {
  const { next, ...pp } = defaultParams;
  await next({
    ...pp,
    ...params
  });
};

// src/error.ts
var runtimeError = (type, msg) => {
  const err = new Error(msg);
  err.name = type;
  return err;
};

// src/client.ts
var ccs = /* @__PURE__ */ new Map();
var fetchStream = (stream, input, codec) => {
  const inputIterator = streamingIterables.transform(
    Infinity,
    (data) => codec.encoder(data),
    input
  );
  const outputIterator = streamingIterables.transform(
    Infinity,
    async (data) => await codec.decoder(data.subarray()),
    stream.source
  );
  stream.sink(inputIterator);
  return outputIterator;
};
var handleControlMsg = async (id, cc) => {
  if (cc === void 0)
    cc = ccs.get(id);
  const msg = (await cc.next()).value;
  cc.close();
  if (msg) {
    try {
      switch (msg.type) {
        case "error":
          const err = new Error(msg.message);
          err.name = msg.name;
          err.stack = msg.stack;
          throw err;
        default:
          break;
      }
    } finally {
      ccs.delete(id);
    }
  }
};
var channel = async (name, connection, options) => {
  const runtimeOptions = {
    ...defaultInitOptions,
    ...options
  };
  logger.trace(`new connecting to protocol ${name}`);
  let inputChannel = new queueable.Channel();
  let stream = await connection.newStream(name);
  let outputIterator = fetchStream(
    stream,
    inputChannel,
    runtimeOptions.codec
  );
  let cc = new queueable.Channel();
  let chan = newChannel(
    inputChannel,
    { connection, stream },
    void 0,
    runtimeOptions.context
  );
  const jid = JSON.stringify(chan.id);
  await chan.send(jid);
  if (name !== control_name)
    ccs.set(jid, cc);
  logger.trace(`protocol ${name} connected`);
  const op = new Proxy({
    ...outputIterator,
    ...chan
  }, {
    get(target, p) {
      if (p === Symbol.asyncIterator) {
        return async function* () {
          for await (const v of target[Symbol.asyncIterator]()) {
            yield v;
          }
          await target.done();
          await handleControlMsg(jid, cc);
        };
      }
      if (p === "next") {
        return async () => {
          const result = await target.next();
          if (result.done === true) {
            await target.done();
            await handleControlMsg(jid, cc);
          }
          return result;
        };
      }
      return target[p];
    }
  });
  return Object.assign(async (value) => {
    await op.send(value);
    return await streamingIterables.collect(op);
  }, op);
};
var client = async (node, peer) => {
  const { isMultiaddr, multiaddr } = await import('@multiformats/multiaddr');
  const { isPeerId } = await import('@libp2p/interface/peer-id');
  const { peerIdFromString } = await import('@libp2p/peer-id');
  let peerToDail;
  if (!(peer instanceof Array)) {
    peer = [peer];
  }
  for (const item of peer) {
    try {
      if (!(isPeerId(item) || isMultiaddr(item))) {
        try {
          const addr = multiaddr(item);
          if (typeof addr.getPeerId() === "string") {
            peerToDail = addr;
          } else {
            throw runtimeError(
              "ConnectError",
              "Invalid addr: no peerid in addr"
            );
          }
        } catch (e) {
          try {
            const peerid = peerIdFromString(item);
            peerToDail = peerid;
          } catch (error) {
            throw runtimeError("ConnectError", "Invalid peerid: " + item);
          }
        }
      } else {
        peerToDail = item;
      }
      logger.trace(`Try to connect to ${peerToDail}`);
      const connection = await node.dial(peerToDail);
      logger.trace(`Succeed connect to ${peerToDail}`);
      const controlChan = await channel(
        control_name,
        connection,
        {
          codec: defaultCodec
        }
      );
      streamingIterables.consume(streamingIterables.transform(
        Infinity,
        async (i) => {
          const msg = JSON.parse(i);
          const cc = ccs.get(JSON.stringify(msg.id));
          await cc?.push(msg);
          logger.trace(
            `Send control msg ${JSON.stringify(msg)} to ${JSON.stringify(msg.id)}`
          );
        },
        controlChan
      ));
      return Object.assign(
        async (name, options) => {
          const runtimeOptions = {
            ...defaultInitOptions,
            ...options
          };
          let chan = await channel(
            name,
            connection,
            runtimeOptions
          );
          return new Proxy(chan, {
            async apply(_, __, argArray) {
              if (!chan.stat.open()) {
                chan = await channel(name, connection, runtimeOptions);
              }
              return await chan(argArray.at(0));
            },
            get(_, p) {
              if (p === "send") {
                return async (v) => {
                  if (!chan.stat.open()) {
                    chan = await channel(
                      name,
                      connection,
                      runtimeOptions
                    );
                  }
                  await chan.send(v);
                };
              }
              return chan[p];
            }
          });
        },
        {
          close: () => connection.close()
        }
      );
    } catch (e) {
      logger.trace(`Skip addr ${item} because of ${e}`);
      continue;
    }
  }
  throw runtimeError("ConnectError", "cannot connect to " + peer);
};
var ccs2 = /* @__PURE__ */ new Map();
var server = async (node) => {
  const makeHandleStream = (codec) => async (name, func) => await node.handle(name, async (incomingData) => {
    const inputIterator = streamingIterables.transform(
      Infinity,
      async (data) => await codec.decoder(data.subarray()),
      incomingData.stream.source
    );
    const outputIterator = streamingIterables.transform(
      Infinity,
      (data) => codec.encoder(data),
      await func(inputIterator, incomingData)
    );
    incomingData.stream.sink(outputIterator);
  });
  const serve = async (name, func, options) => {
    const runtimeOptions = {
      ...defaultInitOptions,
      ...options
    };
    await makeHandleStream(runtimeOptions.codec)(
      name,
      async (input, incomingData) => {
        const outputChannel = new queueable.Channel();
        let chan;
        for await (const id of input) {
          const sid = JSON.parse(id);
          incomingData.connection.id = sid.connection;
          incomingData.stream.id = sid.stream;
          const cc = ccs2.get(sid.connection);
          if (cc === void 0 && name !== control_name) {
            throw runtimeError("ChannelNotFound", "Control channel not found");
          }
          chan = newChannel(
            outputChannel,
            incomingData,
            cc,
            runtimeOptions.context
          );
          logger.trace(`New connection with ${id}`);
          break;
        }
        try {
          const process = await func(chan) ?? (() => {
          });
          streamingIterables.consume(
            streamingIterables.transform(
              Infinity,
              async (data) => {
                logger.trace(`Incoming data: ${JSON.stringify(data)}`);
                try {
                  if (runtimeOptions.middleware) {
                    await runtimeOptions.middleware({
                      data,
                      ...chan,
                      next: makeNext({ data, ...chan, next: process })
                    });
                  } else {
                    await process({ data, ...chan });
                  }
                } catch (error) {
                  await chan.done(error);
                }
              },
              input
            )
          ).then(() => chan?.done());
        } catch (error) {
          await chan?.done(error);
        }
        return outputChannel;
      }
    );
  };
  const handle = async (name, func, options) => await serve(name, () => func, options);
  if (!node.getProtocols().some((p) => p === control_name)) {
    await serve(control_name, (chan) => {
      const id = chan.id.connection;
      const cc = new queueable.Channel();
      ccs2.set(id, cc);
      streamingIterables.consume(
        streamingIterables.transform(Infinity, async (i) => {
          try {
            await chan.send(JSON.stringify(i));
          } catch (error) {
            if (error === "channel has already closed") {
              await cc.return();
              ccs2.delete(id);
            } else
              throw error;
          }
        }, cc)
      );
    }, {
      codec: defaultCodec
    });
  }
  return { handle, serve };
};

exports.client = client;
exports.server = server;
